"""Module for geometric operations."""

import numpy as np
from numba import njit


@njit
def normalize(vec):
    """Normalize a vector.

    Parameters
    ----------
    vec : numpy 1D-array

    Returns
    -------
    numpy 1D-array
        The normalized vector.
    """
    return vec / norm(vec)


@njit
def norm(vec):
    """Return the norm of a vector.

    Parameters
    ----------
    vec : numpy 1D-array

    Returns
    -------
    float
        The magniture of the vector.
    """
    return np.sqrt((vec**2).sum())


@njit
def calc_angle(atom1, atom2, atom3):
    """Calculate the valence angle between atom1, atom2 and atom3.

    Note
    ----
    `atom2` is the central atom.

    Parameters
    ----------
    atom1 : numpy 1D-array.
    atom2 : numpy 1D-array.
    atom3 : numpy 1D-array.

    Returns
    -------
    float
        The calculated angle in radians.
    """
    vec1 = atom1 - atom2
    vec2 = atom3 - atom2
    costheta = np.dot(vec1, vec2)/(norm(vec1)*norm(vec2))
    if costheta > 1.0 or costheta < -1.0:
        raise ValueError("Cosine cannot be larger than 1.0 or less than -1.0")
    return np.arccos(costheta)


@njit
def vec2quaternion(vec, theta):
    """Translate a vector of 3 elements and angle theta to a quaternion.

    Parameters
    ----------
    vec : numpy 1D-array
        Vector of the quaternion.
    theta : float
        Angle of the quaternion in radian.

    Returns
    -------
    numpy 1D-array
        The full quaternion (4 elements).
    """
    w = np.cos(theta/2)
    x, y, z = np.sin(theta/2) * normalize(vec)
    return np.array([w, x, y, z])


@njit
def calc_rotation_matrix(quaternion):
    """Translate a quaternion to a rotation matrix.

    Parameters
    ----------
    quaternion : numpy 1D-array of 4 elements.

    Returns
    -------
    numpy 2D-array (dimension `[3, 3]`)
        The rotation matrix.
    """
    # Initialize rotation matrix.
    matrix = np.zeros((3, 3), dtype=np.float32)
    # Get quaternion elements.
    w, x, y, z = quaternion
    # Compute rotation matrix.
    matrix[0, 0] = w**2 + x**2 - y**2 - z**2
    matrix[1, 0] = 2 * (x*y + w*z)
    matrix[2, 0] = 2 * (x*z - w*y)
    matrix[0, 1] = 2 * (x*y - w*z)
    matrix[1, 1] = w**2 - x**2 + y**2 - z**2
    matrix[2, 1] = 2 * (y*z + w*x)
    matrix[0, 2] = 2 * (x*z + w*y)
    matrix[1, 2] = 2 * (y*z - w*x)
    matrix[2, 2] = w**2 - x**2 - y**2 + z**2
    return matrix


@njit
def apply_rotation(vec_to_rotate, rotation_axis, rad_angle):
    """Rotate a vector around an axis by a given angle.

    Note
    ----
    The rotation axis is a vector of 3 elements.

    Parameters
    ----------
    vec_to_rotate : numpy 1D-array
    rotation_axis : numpy 1D-array
    rad_angle : float

    Returns
    -------
    numpy 1D-array
        The final rotated (normalized) vector.
    """
    # Generate a quaternion of the given angle (in radian).
    quaternion = vec2quaternion(rotation_axis, rad_angle)
    # Generate the rotation matrix.
    rotation_matrix = calc_rotation_matrix(quaternion)
    # Apply the rotation matrix on the vector to rotate.
    vec_rotated = np.dot(rotation_matrix, vec_to_rotate)
    return normalize(vec_rotated)


@njit
def cross_product(A, B):
    """Return the cross product between vectors A & B.

    Source: http://hyperphysics.phy-astr.gsu.edu/hbase/vvec.html.

    Note
    ----
    On small vectors (i.e. of 3 elements), computing cross products
    with this functions is faster than `np.cross()`.

    Parameters
    ----------
    A : numpy 1D-array
        A vector of 3 elements.
    B : numpy 1D-array
        Another vector of 3 elements.

    Returns
    -------
    numpy 1D-array
        Cross product of A^B.
    """
    x = (A[1]*B[2]) - (A[2]*B[1])
    y = (A[0]*B[2]) - (A[2]*B[0])
    z = (A[0]*B[1]) - (A[1]*B[0])
    return np.array((x, -y, z))


@njit
def calc_OP(C, H):
    """Return the Order Parameter of a CH bond (OP).

    OP is calculated according to equation:

    .. math:: S = 1/2 * (3*cos(theta)^2 -1)

    theta is the angle between CH bond and the z(vertical) axis:

    .. code::

        z
        ^   H
        |  /
        | /
        C

    Inspired from a function written by @jmelcr.

    Parameters
    ----------
    C : numpy 1D-array
        Coordinates of C atom.
    H : numpy 1D-array
        Coordinates of H atom.

    Returns
    -------
    float
        The normalized vector.
    """
    vec = H - C
    d2 = np.square(vec).sum()
    cos2 = vec[2]**2/d2
    S = 0.5*(3.0*cos2 - 1.0)
    return S



"""User Interface module.

It contains functions used for the CLI mode
and the one for the 'module mode'.
"""

import argparse
import pickle
import pathlib
import sys

import numpy as np
import MDAnalysis as mda

from . import __version__
from . import lipids
from . import init_dics
from . import core
from . import writers
from . import utils


class BuildHError(Exception):
    """Raised when BuildH returns a error.

    This should be used when using BuildH as a module.
    """

    pass


def isfile(path):
    """Callback for checking file existence.

    This function checks if `path` is an existing file.
    If not, raise an error. Else, return the `path`.

    Parameters
    ----------
    path : str
        The path to be checked.

    Returns
    -------
    str
        The validated path.
    """
    source = pathlib.Path(path)
    if not pathlib.Path.is_file(source):
        if pathlib.Path.is_dir(source):
            msg = f"{source} is a directory."
        else:
            msg = f"{source} does not exist."
        raise argparse.ArgumentTypeError(msg)
    return path


def parse_cli():
    """Handle the user parameters from the command line."""
    # Retrieve list of supported lipids
    lipids_files = [f for f in lipids.PATH_JSON.iterdir() if f.is_file()]
    lipids_tops = lipids.read_lipids_topH(lipids_files)
    lipids_supported_str = ", ".join(lipids_tops.keys())

    message = """This program builds hydrogens and calculates the order
    parameters (OP) from a united-atom trajectory of lipids. If -opx is
    requested, pdb and xtc output files with hydrogens are created but
    OP calculation will be slow. If no trajectory output is requested
    (no use of flag -opx), it uses a fast procedure to build hydrogens
    and calculate the OP.
    """
    epilog = f"The list of supported lipids (-l option) are: {lipids_supported_str}. "
    epilog += "More documentation can be found at https://buildh.readthedocs.io."
    parser = argparse.ArgumentParser(description=message,
                                     epilog=epilog)
    parser.add_argument('-v', '--version', action='version', version=__version__)
    # Avoid tpr for topology cause there's no .coord there!
    parser.add_argument("-c", "--coord", type=isfile, required=True,
                        help="Coordinate file (pdb or gro format).")
    parser.add_argument("-t", "--traj", type=isfile,
                        help="Input trajectory file. Could be in XTC, TRR or DCD format.")
    parser.add_argument("-l", "--lipid", type=str, required=True,
                        help="Combinaison of ForceField name and residue name "
                        "for the lipid to calculate the OP on (e.g. Berger_POPC)."
                        "It must match with the internal topology files or the one(s) supplied."
                        "A list of supported terms is printed when calling the help.")
    parser.add_argument("-lt", "--lipid_topology", type=isfile, nargs='+',
                        help="User topology lipid json file(s).")
    parser.add_argument("-d", "--defop", required=True, type=isfile,
                        help="Order parameter definition file. Can be found on "
                        "https://github.com/patrickfuchs/buildH/tree/master/def_files.")
    parser.add_argument("-opx", "--opdbxtc", help="Base name for trajectory "
                        "output with hydrogens. File extension will be "
                        "automatically added. For example -opx trajH will "
                        "generate trajH.pdb and trajH.xtc. "
                        "So far only xtc is supported.")
    parser.add_argument("-o", "--out", help="Output file name for storing "
                        "order parameters. Default name is OP_buildH.out.",
                        default="OP_buildH.out")
    parser.add_argument("-b", "--begin", type=int,
                        help="The first frame (ps) to read from the trajectory.")
    parser.add_argument("-e", "--end", type=int,
                        help="The last frame (ps) to read from the trajectory.")
    # parser.add_argument("-pi", "--pickle", type=str,
    #                     help="Output pickle filename. The structure pickled is a dictonnary "
    #                     "containing for each Order parameter, "
    #                     "the value of each lipid and each frame as a matrix")
    options = parser.parse_args()


    # Check coord file extension.
    if not options.coord.endswith("pdb") and not options.coord.endswith("gro"):
        parser.error("Coordinates file must be given in .pdb or .gro format.")

    # Use only the user lipid topologies
    if options.lipid_topology:
        lipids_files = [pathlib.Path(f) for f in options.lipid_topology]
        # Regenerate lipid topologies dictionary
        try:
            lipids_tops = lipids.read_lipids_topH(lipids_files)
        except ValueError as e:
            parser.error(e)
        # Regenerate str list of supported lipids.
        lipids_supported_str = ", ".join(lipids_tops.keys())

    # Check residue name validity.
    # Get the dictionary with helper info using residue name (options.lipid
    # argument).
    try:
        lipids_info = lipids_tops[options.lipid]
    except KeyError:
        parser.error(f"Lipid {options.lipid} is not supported. "
                     f"List of supported lipids are: {lipids_supported_str}")

    # Slicing only makes sense with a trajectory
    if not options.traj and (options.begin or options.end):
        parser.error("Slicing is only possible with a trajectory file.")

    return options, lipids_info


def entry_point():
    """Correspond to the entry point `buildH`."""
    # Parse arguments.
    args, dic_lipid = parse_cli()

    try:
        main(args.coord, args.traj, args.defop, args.out, args.opdbxtc, dic_lipid, args.begin, args.end)
    except BuildHError as e:
        sys.exit(e)


def launch(coord_file, def_file, lipid_type, traj_file=None, out_file="OP_buildH.out", prefix_traj_ouput=None, begin=None, end=None, lipid_jsons=None):
    """Launch BuildH calculations.

    This is the only function which can be called inside a Python script to use BuildH as a module.
    It checks the different arguments and call the main function.

    Parameters
    ----------
    coord_file : str
        Coordinate file. Only .pdb and .gro files are currently supported.
    def_file : str
        Order parameter definition file.
    lipid_type : str
        Combinaison of ForceField name and residue name for the lipid to calculate the OP on (e.g. Berger_POPC).
        It must match with the internal topology files or the one(s) supplied.
    traj_file : str, optional
        Trajectory file (could be in XTC, TRR or DCD format), by default None.
    out_file : str, optional
        Output file name for storing order parameters, by default "OP_buildH.out".
    prefix_traj_ouput : str, optional
        Base name for trajectory output with hydrogens.
        File extension will be automatically added.
        By default None.
    begin : int, optional
        The first frame (ps) to read from the trajectory, by default None.
    end : int, optional
        The last frame (ps) to read from the trajectory, by default None.
    lipid_jsons : list, optional
        User topology lipid json file(s), by default None.

    Raises
    ------
    FileNotFoundError
        When either coord_file, def_file or the traj_file is missing.
    TypeError
        When lipid_jsons is not a list.
    BuildHError
        When something went wront during calculation.
    """
    # Check files
    for file in [coord_file, traj_file, def_file]:
        if file is not None:
            source = pathlib.Path(file)
            if not pathlib.Path.is_file(source):
                raise FileNotFoundError(f"{source} does not exist.")

    # Construct available lipid topologies with only the user topology files.
    if lipid_jsons:
        if not isinstance(lipid_jsons, list):
            raise TypeError(f"a list is expected for argument 'lipid_jsons'")
        lipids_files = [pathlib.Path(f) for f in lipid_jsons]
        # Generate lipid topologies dictionary
        try:
            lipids_tops = lipids.read_lipids_topH(lipids_files)
        except ValueError as e:
            raise BuildHError(e)
    # Generate lipid topologies dictionary from internal topology files.
    else:
        lipids_files = [f for f in lipids.PATH_JSON.iterdir() if f.is_file()]
        lipids_tops = lipids.read_lipids_topH(lipids_files)

    try:
        dic_lipid = lipids_tops[lipid_type]
    except KeyError:
        raise BuildHError(f"Lipid {lipid_type} is not supported. "
                               f"List of supported lipids are: " + ", ".join(lipids_tops.keys()))


    try:
        main(coord_file, traj_file, def_file, out_file, prefix_traj_ouput, dic_lipid, begin, end)
    except BuildHError as e:
        raise e


def main(coord_file, traj_file, def_file, out_file, prefix_traj_ouput, dic_lipid, begin=None, end=None):
    """Main function of BuildH.

    It takes care of all the necessary steps to compute the Order Parameter :
      - create MDAnalysis Universe
      - create internal dictionaries
      - Check topology
      - Reconstruct the hydrogens
      - Compute OP and write the result in the output file.
      - If asked, write the trajectory with the hydrogens.

    Note
    ----
    This function shouldn't be called directly.
    Either it's called by the entry point or by the `launch` function.

    Parameters
    ----------
    coord_file : str
        Coordinate file. Only .pdb and .gro files are currently supported.
    traj_file : str
        Trajectory file (could be in XTC, TRR or DCD format). Can be None.
    def_file : str
        Order parameter definition file.
    out_file : str
        Output file name for storing order parameters.
    prefix_traj_ouput : str
        Base name for trajectory output with hydrogens.
        File extension will be automatically added.
        Can be None.
    dic_lipid : dict
        Lipid Topology for the reconstruction of the hydrogens.
    begin : int
        The first frame (ps) to read from the trajectory, by default None.
    end : int
        The last frame (ps) to read from the trajectory, by default None.

    Raises
    ------
    BuildHError
        When something went wront during calculation.
    """
    # Create universe without H.
    print("Constructing the system...")
    if traj_file:
        try:
            universe_woH = mda.Universe(coord_file, traj_file)
            begin, end = utils.check_slice_options(universe_woH, begin, end)
            traj_file = True
        except IndexError:
            raise BuildHError("Slicing options are not correct.")
        except:
            raise BuildHError(f"Can't create MDAnalysis universe with files {coord_file} and {traj_file}.")
    else:
        try:
            universe_woH = mda.Universe(coord_file)
            begin = 0
            end = 1
            traj_file = False
        except:
            raise BuildHError(f"Can't create MDAnalysis universe with file {coord_file}.")


    # Initialize dic for storing OP.
    # Init dic of correspondance : {('C1', 'H11'): 'gamma1_1',
    # {('C1', 'H11'): 'gamma1_1', ...}.
    try:
        dic_atname2genericname = init_dics.make_dic_atname2genericname(def_file)
    except ValueError as e:
        raise BuildHError(e)
    # Initialize dic_OP (see function init_dic_OP() for the format).
    dic_OP, dic_corresp_numres_index_dic_OP = init_dics.init_dic_OP(universe_woH,
                                                                    dic_atname2genericname,
                                                                    dic_lipid['resname'])
    # Initialize dic_Cname2Hnames.
    dic_Cname2Hnames = init_dics.make_dic_Cname2Hnames(dic_OP)


    # Now, a few checks have to be performed to ensure the lipid topology chosen (-l option),
    # the structure provided and the def file provider are coherent with each others.

    # Check if the lipid topology match the the structure.
    if not lipids.check_topology(universe_woH, dic_lipid):
        raise BuildHError(f"The topology chosen does not match the structure provided in {coord_file}")

    # Check if atoms name in the def file are present in the structure.
    atoms_name = [heavy_atom for (heavy_atom, _) in dic_atname2genericname.keys()]
    if not utils.check_def_file(universe_woH, dic_lipid['resname'], atoms_name):
        raise BuildHError(f"Atoms defined in {def_file} are missing in the structure {coord_file}.")

    # Check the def file and the topology are coherent.
    if not utils.check_def_topol_consistency(dic_Cname2Hnames, dic_lipid):
        raise BuildHError(f"Atoms defined in {def_file} are not consistent with the chosen topology.")


    print("System has {} atoms".format(len(universe_woH.coord)))

    # If traj output files are requested.
    # NOTE Here, we need to reconstruct all Hs. Thus the op definition file (passed
    #  with arg -d) needs to contain all possible C-H pairs !!!
    if prefix_traj_ouput:

        if utils.is_allHs_present(def_file, dic_lipid, dic_Cname2Hnames):
            core.gen_coordinates_calcOP(prefix_traj_ouput, universe_woH, dic_OP, dic_lipid,
                                        dic_Cname2Hnames, dic_corresp_numres_index_dic_OP,
                                        begin, end, traj_file)
        else:
            raise BuildHError(f"Error on the number of H's to rebuild. An output trajectory has been "
                     f"requestest but {def_file} doesn't contain all hydrogens to rebuild.")

    # If no traj output file requested, use fast indexing to speed up OP
    # calculation. The function fast_build_all_Hs() returns nothing, dic_OP
    # is modified in place.
    if not prefix_traj_ouput:
        core.fast_build_all_Hs_calc_OP(universe_woH, begin, end, dic_OP,
                                       dic_lipid, dic_Cname2Hnames)


    # Output to a file.
    writers.write_OP(out_file, dic_atname2genericname,
                            dic_OP, dic_lipid['resname'])
    print(f"Results written to {out_file}")

    # # Pickle results
    # if args.pickle:
    #     with open(args.pickle, "wb") as f:
    #         # Pickle the dic using the highest protocol available.
    #         pickle.dump(dic_OP, f, pickle.HIGHEST_PROTOCOL)
    #         print("dictionary pickled and written to {}".format(args.pickle))
    #     #  To unpickle
    #     #with open("OP.pickle", "rb") as f:
    #     #    dic_OP = pickle.load(f)



"""Module holding the core functions."""

import numpy as np
import pandas as pd
import MDAnalysis as mda
import MDAnalysis.coordinates.XTC as XTC

from . import hydrogens
from . import geometry as geo
from . import writers


# For debugging.
# TODO: Remove it after logging feature implementation.
DEBUG=False


def buildHs_on_1C(atom, H_type, helper1, helper2, helper3=None):
    """Build 1, 2 or 3 H on a given carbon.

    This function is a wrapper which takes the coordinates of the helpers
    and call the function that builds 1, 2 or 3 H.

    Parameters
    ----------
    atom : numpy 1D-array
        Central atom on which we want to reconstruct the hydrogen.
    H_type: str
        The type of H to build. It could be 'CH2', 'CH', 'CHdoublebond' or 'CH3'
        (see `init_dics.py`)
    helper1 : numpy 1D-array
        First neighbor of central atom.
    helper2 : numpy 1D-array
        Second neighbor of central atom.
    helper3 : numpy 1D-array
        Third neighbor of central atom.

    Returns
    -------
    tuple of numpy 1D-arrays
        Each element of the tuple is a numpy 1D-array containing 1, 2 or 3
        reconstructed hydrogen(s).
        **Important**: This function returns a tuple even if
        there's only one H that has been rebuilt.
    """
    if H_type == "CH2":
        H1_coor, H2_coor = hydrogens.get_CH2(atom, helper1, helper2)
        return (H1_coor, H2_coor)
    elif H_type == "CH":
        # If we reconstruct a single H, we have a 3rd helper.
        #helper3_coor = sel("name {0}".format(helper3_name))[0].position
        H1_coor = hydrogens.get_CH(atom, helper1, helper2,
                                   helper3)
        return (H1_coor,)
    elif H_type == "CHdoublebond":
        H1_coor = hydrogens.get_CH_double_bond(atom, helper1,
                                               helper2)
        return (H1_coor,)
    elif H_type == "CH3":
        H1_coor, H2_coor, H3_coor = hydrogens.get_CH3(atom,
                                                      helper1, helper2)
        return (H1_coor, H2_coor, H3_coor)
    else:
        raise UserWarning("Wrong code for typeofH2build, expected 'CH2', 'CH'"
                          ", 'CHdoublebond' or 'CH3', got {}."
                          .format(H_type))


def build_system_hydrogens(universe_woH, dic_lipid, dic_Cname2Hnames, dic_lipid_indexes):
    """Build a new system **with** hydrogens.

    The function takes the MDAnalysis universe **without** hydrogens, reconstructs all hydrogens
    and returns a pandas dataframe. This latter will be used later to build a new
    MDAnalysis universe with H.

    Note
    ----
    There is no simple way to create a new MDAnalysis universe directly.

    Parameters
    ----------
    universe_woH : MDAnalysis universe instance
        This is the universe **without** hydrogen.
    dic_lipid : dictionary
        Contains carbon names and helper names needed for reconstructing hydrogens.
    dic_Cname2Hnames : dictionary
        This dict gives the correspondance Cname -> Hname.
        (see `make_dic_Cname2Hnames()` in `init_dics.py`)
    dic_lipids_with_indexes : dictionary
        The dictionary made in function `make_dic_lipids_with_indexes()`.

    Returns
    -------
    pandas dataframe
        contains the system **with** hydrogens.
    """
    # The list newrows will be used to store the new molecule *with* H.
    newrows = []
    # Counter for numbering the new mlcs with H.
    new_atom_num = 1
    # Loop over all atoms in the universe without H.
    for atom in universe_woH.atoms:
        # Make resnum resets to 0 when it exceeds 9999.
        resnum = atom.resnum-((atom.resnum//10000)*10000)
        resname = atom.resname
        name = atom.name
        # Append atom to the new list.
        # 0      1       2        3       4  5  6
        # atnum, atname, resname, resnum, x, y, z
        newrows.append([new_atom_num, name, resname, resnum]
                        + list(atom.position))
        new_atom_num += 1
        # Build new H(s)?
        if (atom.name in dic_lipid and atom.residue.resname == dic_lipid["resname"]):
            # Retrieve helpers coordinates
            # helperX_ix is the index of the helper inside one residue.
            if len(dic_lipid_indexes[atom.name]) == 6:
                typeofH2build, _, _, _, helper1_ix, helper2_ix = dic_lipid_indexes[atom.name]
                helper3_coor = None
            else:
                typeofH2build, _, _, _, _, helper1_ix, helper2_ix, helper3_ix = dic_lipid_indexes[atom.name]
                helper3_coor = atom.residue.atoms[helper3_ix].position

            helper1_coor = atom.residue.atoms[helper1_ix].position
            helper2_coor = atom.residue.atoms[helper2_ix].position

            # Build Hs and store them in a list of numpy 1D-arrays Hs_coor.
            # The "s" in Hs_coor means there can be more than 1 H:
            # For CH2, Hs_coor will contain: [H1_coor, H2_coor].
            # For CH3, Hs_coor will contain: [H1_coor, H2_coor, H3_coor].
            # For CH, Hs_coor will contain: [H1_coor].
            # For CHdoublebond, Hs_coor will contain: [H1_coor].
            Hs_coor = buildHs_on_1C(atom.position, typeofH2build,
                                    helper1_coor, helper2_coor, helper3_coor)

            # Loop over Hs_coor (H_coor is a 1D-array with the 3 coors of 1 H).
            for i, H_coor in enumerate(Hs_coor):
                # Retrieve name of newly built H.
                Hname = dic_Cname2Hnames[atom.name][i]
                # Add them to newrows.
                newrows.append([new_atom_num, Hname, resname, resnum]
                                + list(H_coor))
                new_atom_num += 1

    # Create a dataframe to store the mlc with added hydrogens.
    new_df_atoms = pd.DataFrame(newrows, columns=["atnum", "atname",
                                                  "resname", "resnum",
                                                  "x", "y", "z"])
    return new_df_atoms


###
### The next function build_all_Hs_calc_OP())
### build new H, calculate the order parameter and write the new traj with Hs
### to an output file (e.g. .xtc, etc).
### Note: it is slow, it shouldn't be used if the user doesn't want to
###       write the trajectory. Instead, fast_build_all_Hs() should be used.
###
def build_all_Hs_calc_OP(universe_woH, ts, dic_lipid, dic_Cname2Hnames, universe_wH, dic_OP,
                         dic_corresp_numres_index_dic_OP, dic_lipid_indexes):
    """Build all hydrogens and calculates order parameters for one frame.

    This function loop overs **all** atoms of the universe_woH in order to update
    the atom coordinates and the new H built into the universe_wH.

    The function also calculates the order parameter.
    The coordinates of the universe **with** H are updated in place.
    The order parameter is also added in place (within dic_OP dictionary).

    Note
    ----
    This function is slow, thus it shall be used when one wants
    to create a trajectory with H (such as .xtc or whatever format).

    This function assumes all possible C-H pairs are present in the .def
    file (with -d option). They are needed since we want to build an xtc with
    the whole system. If one is interested in calculating only a subset of OPs,
    please use the function `fast_build_all_Hs_calc_OP()` instead.

    Parameters
    ----------
    universe_woH : MDAnalysis universe instance
        This is the universe **without** hydrogen.
    ts : Timestep instance
        the current timestep with the coordinates
    dic_lipid : dictionary
        Contains carbon names and helper names needed for reconstructing hydrogens.
    dic_Cname2Hnames : dictionary
        This dict gives the correspondance Cname -> Hname.
        (see `make_dic_Cname2Hnames()` in `init_dics.py`)
    universe_wH : MDAnalysis universe instance (optional)
        This is the universe **with*** hydrogens.
    dic_OP : ordered dictionary
        Each key of this dict is a couple carbon/H, and at the beginning it
        contains an empty list.
        (see `init_dic_OP()` in `init_dics.py`)
    dic_corresp_numres_index_dic_OP : dictionary
        This dict should contain the correspondance between the numres and
        the corresponding index in `dic_OP`. For example `{15: 14}` means
        the residue numbered 15 in the PDB has an index of 14 in `dic_OP`.
    """
    # We will need the index in the numpy array for updating coordinates
    # in the universe with H.
    row_index_coor_array = 0
    resid = -9999

    # Loop over all atoms in the universe without H.
    for atom in universe_woH.atoms:
        # Update the position of the current atom in the universe with H.
        universe_wH.coord.positions[row_index_coor_array, :] = atom.position
        row_index_coor_array += 1

        # Build new H(s)?
        if (atom.name in dic_lipid and atom.residue.resname == dic_lipid["resname"]):

            # Retrieve the index of the first atom in the current residue
            # Test to avoid refreshing it at every step of the loop
            if resid != atom.residue.resid:
                resid = atom.residue.resid
                ix_first_atom_res = atom.residue.atoms[0].ix

            # Retrieve helpers coordinates
            if len(dic_lipid_indexes[atom.name]) == 6:
                typeofH2build, _, _, _, helper1_ix, helper2_ix = dic_lipid_indexes[atom.name]
                helper3_coor = None
            else:
                typeofH2build, _, _, _, _, helper1_ix, helper2_ix, helper3_ix = dic_lipid_indexes[atom.name]
                helper3_coor = ts[helper3_ix + ix_first_atom_res]

            # Faster to retrieve the coordinates from ts than from universe_woH.atoms.positions
            helper1_coor = ts[helper1_ix + ix_first_atom_res]
            helper2_coor = ts[helper2_ix + ix_first_atom_res]

            # Build Hs and store them in a list of numpy 1D-arrays Hs_coor.
            # The "s" in Hs_coor means there can be more than 1 H:
            # For CH2, Hs_coor will contain: [H1_coor, H2_coor].
            # For CH3, Hs_coor will contain: [H1_coor, H2_coor, H3_coor].
            # For CH, Hs_coor will contain: [H1_coor].
            # For CHdoublebond, Hs_coor will contain: [H1_coor].
            Hs_coor = buildHs_on_1C(atom.position, typeofH2build,
                                    helper1_coor, helper2_coor, helper3_coor)

            # Loop over Hs_coor (H_coor is a 1D-array with the 3 coors of 1 H).
            for i, H_coor in enumerate(Hs_coor):
                # Retrieve name of newly built H.
                Hname = dic_Cname2Hnames[atom.name][i]
                ####
                #### We calculate here the order param on the fly :-D !
                ####
                if (atom.name, Hname) in dic_OP:
                    op = geo.calc_OP(atom.position, H_coor)
                    # We should get here the index of the residue in dic_OP.
                    # For that we can use dic_corresp_numres_index_dic_OP
                    # (key: resnum in pdb, value: index residue in dic_OP).
                    lipid_ix = dic_corresp_numres_index_dic_OP[atom.resid]
                    # OLD way: dic_OP[(atom.name, Hname)].append(op)
                    if (atom.name, Hname) in dic_OP:
                        dic_OP[(atom.name, Hname)][lipid_ix].append(op)
                    if DEBUG:
                        print(atom.name, H_coor, "OP:", op)

                # Update the position of the current H in the universe with H.
                universe_wH.coord.positions[row_index_coor_array, :] = H_coor
                row_index_coor_array += 1

            if dic_OP and DEBUG:
                print()
                print()
    if dic_OP and DEBUG:
        print("Final dic_OP:", dic_OP)
        print()


###
### The next 3 functions (get_indexes(), make_dic_lipids_with_indexes()
### and fast_build_all_Hs_calc_OP()) should be used when the
### user doesn't want an output trajectory.
### By using fast indexing to individual Catoms and helpers, they
### are much faster.
###
def get_indexes(atom, dic_lipid):
    """Return the index of helpers for a given carbon.

    Parameters
    ----------
    atom : MDAnalysis Atom instance
        This is an Atom instance of a carbon on which we want to build Hs.
    dic_lipid : dictionary
        Contains carbon names and helper names needed for reconstructing hydrogens.

    Returns
    -------
    tuple of 2 or 3 int
        The tuple contains the index of the 2 (or 3) helpers for the atom that
        was passed as argument. (e.g. for atom C37 with index 99, the function
        returns a tuple containing 98 (index of C36 = helper 1) and 100 (index
        of C38=helper2).
    """
    # Get nb of H to build and helper names (we can have 2 or 3 helpers).
    if len(dic_lipid[atom.name]) == 3:
        typeofH2build, helper1_name, helper2_name = dic_lipid[atom.name]
    else:
        typeofH2build, helper1_name, helper2_name, helper3_name = dic_lipid[atom.name]
    # Get helper coordinates using atom, which an instance from Atom class.
    # atom.residue.atoms is a list of atoms we can select with
    # method .select_atoms().
    # To avoid too long line, we shorten its name to `sel`.
    sel = atom.residue.atoms.select_atoms
    helper1_ix = sel("name {}".format(helper1_name))[0].ix
    helper2_ix = sel("name {}".format(helper2_name))[0].ix
    if typeofH2build == "CH":
        # If we reconstruct a single H, we have a 3rd helper.
        helper3_ix = sel("name {0}".format(helper3_name))[0].ix
        return (helper1_ix, helper2_ix, helper3_ix)
    else:
        return (helper1_ix, helper2_ix)


def make_dic_lipids_with_indexes(universe_woH, dic_lipid, dic_OP):
    """Expand dic_lipid and adds the index of each atom and helper.

    **IMPORTANT**: the index of each atom/helper is given with respect to the first atom in that residue.

    For example, if we have a POPC where C1 is the first atom, and C50 the
    last one, we want in the end:

    .. code::

        {'C1': ('CH3', 'N4', 'C5', 0, 3, 4), ...,
         'C50': ('CH3', 'C49', 'C48', 49, 48, 47)}

    Where the 3 last int are the index (ix) of the atom, helper1, helper2
    (possibly helper3) with respect to the first atom.
    Thus for C1 : 0 is index of C1, N4 is 3 atoms away from C1 and C5 is 4
    atoms away from C1.
    For C50: C50 is 49 atoms away from C1, C49 is 48 atoms away from C1,
    C48 is 47 atoms away from C1.

    Parameters
    ----------
    universe_woH : MDAnalysis Universe instance
        The universe **without** hydrogens.
    dic_lipid : dictionary
        Contains carbon names and helper names needed for reconstructing hydrogens.
    dic_OP : ordered dictionary
        Each key of this dict is a couple carbon/H, and at the beginning it
        contains an empty list.
        (see `init_dic_OP()` in `init_dics.py`)

    Returns
    -------
    dictionary
        The returned dictionary as described above in this docstring.
    """
    # Get lipid name.
    resname = dic_lipid["resname"]
    # Get resnum of the 1st lipid encountered in the system whose name
    # is `resname`.
    selection = "resname {}".format(resname)
    first_lipid_residue = universe_woH.select_atoms(selection).residues[0]
    resnum_1st_lipid = first_lipid_residue.resnum
    # Get name of 1st atom of that lipid.
    first_atom_name = first_lipid_residue.atoms[0].name
    # Get index of this atom.
    first_atom_ix = first_lipid_residue.atoms[0].ix
    if DEBUG:
        print("resname: {}, first encountered residue: {},\n"
              "resnum_1st_lipid: {}, first_atom_name: {}, first_atom_ix: {}"
              .format(resname, first_lipid_residue, resnum_1st_lipid,
                      first_atom_name, first_atom_ix))
        print()
    # Keep only carbons on which we want to build Hs.
    carbons2keep = []
    for Cname, _ in dic_OP:
        if Cname not in carbons2keep:
            carbons2keep.append(Cname)
    dic_lipids_with_indexes = {}
    for Cname in dic_lipid.keys():
        if Cname in carbons2keep:
            dic_lipids_with_indexes[Cname] = dic_lipid[Cname].copy()
    # Now add the helper indexes.
    # The reasonning is over one residue (e.g. POPC). We want to add (to the
    # dict) the index (ix) of each helper of a given carbon with respect to
    # the index of the first atom in that lipid residue.
    # Loop over each carbon on which we want to reconstruct Hs.
    for Cname in dic_lipids_with_indexes:
        # Loop over residues for a given Cname atom.
        selection = "resid {} and name {}".format(resnum_1st_lipid, Cname)
        for Catom in universe_woH.select_atoms(selection):
            # Get the (absolute) index of helpers.
            if dic_lipid[Cname][0] == "CH":
                helper1_ix, helper2_ix, helper3_ix = get_indexes(Catom, dic_lipid)
            else:
                helper1_ix, helper2_ix = get_indexes(Catom, dic_lipid)
            # If the first lipid doesn't start at residue 1 we must
            # substract the index of the first atom of that lipid.
            Catom_ix_inres = Catom.ix - first_atom_ix
            helper1_ix_inres = helper1_ix - first_atom_ix
            helper2_ix_inres = helper2_ix - first_atom_ix
            # Then add these indexes to dic_lipids_with_indexes.
            if dic_lipid[Cname][0] == "CH":
                helper3_ix_inres = helper3_ix - first_atom_ix
                tmp_tuple = (Catom_ix_inres, helper1_ix_inres,
                             helper2_ix_inres, helper3_ix_inres)
                dic_lipids_with_indexes[Cname] += tmp_tuple
            else:
                tmp_tuple = (Catom_ix_inres, helper1_ix_inres,
                             helper2_ix_inres)
                dic_lipids_with_indexes[Cname] += tmp_tuple
    if DEBUG:
        print("Everything is based on the following dic_lipids_with_indexes\n{}"
              .format(dic_lipids_with_indexes))
        print()
    return dic_lipids_with_indexes


def fast_build_all_Hs_calc_OP(universe_woH, begin, end,
                              dic_OP, dic_lipid, dic_Cname2Hnames):
    """Build Hs and calc OP using fast indexing.

    This function uses fast indexing to carbon atoms and helper atoms. It
    should be used when the user doesn't want any output traj with hydrogens.

    Parameters
    ----------
    universe_woH : MDAnalysis universe instance
        This is the universe **without** hydrogen.
    begin: int
        index of the first frame of trajectory
    end: int
        index of the last frame of trajectory
    dic_OP : ordered dictionary
        Each key of this dict is a couple carbon/H, and at the beginning it
        contains an empty list.
        (see `init_dic_OP()` in `init_dics.py`)
    dic_lipid : dictionary
        Contains carbon names and helper names needed for reconstructing hydrogens.
    dic_Cname2Hnames : dictionary
        This dict gives the correspondance Cname -> Hname.
        (see `make_dic_Cname2Hnames()` in `init_dics.py`)

    Returns
    -------
    None
        This function returns nothing, `dic_OP` is changed **in place**.
    """
    ###
    ### 1) Expand dic_lipids and store there helpers' index.
    ###
    ### We want {'C1': ('CH3', 'N4', 'C5', 0, 3, 4), ...,
    ###          'C50': ('CH3', 'C49', 'C48', 49, 48, 47)}
    ### Where the 3 last int are the index (ix) of the atom, helper1, helper2
    ### (possibly helper3) with respect to the first atom
    ### (e.g. 0 is index of C1, N4 is 3 atoms away from C1, etc).
    ###
    dic_lipids_with_indexes = make_dic_lipids_with_indexes(universe_woH,
                                                           dic_lipid, dic_OP)
    # Get lipid name.
    resname = dic_lipid["resname"]
    # Select first residue of that lipid.
    selection = "resname {}".format(resname)
    first_lipid_residue = universe_woH.select_atoms(selection).residues[0]
    # Get name of 1st atom of that lipid.
    first_atom_name = first_lipid_residue.atoms[0].name
    ###
    ### 2) Now loop over the traj, residues and Catoms.
    ### At each iteration build Hs and calc OP.
    ###
    # Loop over frames (ts is a Timestep instance).
    for ts in universe_woH.trajectory[begin:end]:
        if len(universe_woH.trajectory) == 1:
            print(f"Dealing with current frame.")
        else:
            print(f"Dealing with frame {ts.frame} at "
                  f"{universe_woH.trajectory.time} ps.")
        if DEBUG:
            print("Looping now over residues...")
            print()
        # Loop over the 1st atom of each lipid, which is equiv to loop *over
        # residues* (first_lipid_atom is an Atom instance, lipid_ix is an int
        # that will be used for storing OPs in dic_OP).
        selection = "resname {} and name {}".format(resname, first_atom_name)
        for lipid_ix, first_lipid_atom in enumerate(universe_woH.select_atoms(selection)):
            if DEBUG:
                print("Dealing with Cname", first_lipid_atom)
                print("which is part of residue", first_lipid_atom.residue)
                print("Now looping over atoms of this residue")
                print()
            # Get the index of this first atom.
            ix_first_atom_res = first_lipid_atom.ix
            # Now loop over each carbon on which we want to build Hs
            # (Cname is a string).
            for Cname in dic_lipids_with_indexes:
                # Get Cname and helpers coords.
                if len(dic_lipids_with_indexes[Cname]) == 6:
                    typeofH2build, _, _, Cname_ix, helper1_ix, helper2_ix = dic_lipids_with_indexes[Cname]
                    helper3_coor = None
                else:
                    typeofH2build, _, _, _, Cname_ix, helper1_ix, helper2_ix, helper3_ix = dic_lipids_with_indexes[Cname]
                    helper3_coor = ts[helper3_ix+ix_first_atom_res]
                helper1_coor = ts[helper1_ix+ix_first_atom_res]
                helper2_coor = ts[helper2_ix+ix_first_atom_res]
                Cname_position = ts[Cname_ix+ix_first_atom_res]
                if DEBUG:
                    print("Dealing with Cname", Cname)
                    sel = first_lipid_atom.residue.atoms.select_atoms
                    Cname_atom = sel("name {}".format(Cname))[0]
                    print(Cname_atom, Cname_atom.position)
                    if len(dic_lipid[Cname]) == 3:
                        _, helper1_name, helper2_name = dic_lipid[Cname]
                    else:
                        _, helper1_name, helper2_name, helper3_name = dic_lipid[Cname]
                    helper1_atom = sel("name {}".format(helper1_name))[0]
                    print("helper1", helper1_atom, helper1_atom.position)
                    helper2_atom = sel("name {}".format(helper2_name))[0]
                    print("helper2", helper2_atom, helper2_atom.position)
                    if len(dic_lipid[Cname]) == 4:
                        helper3_atom = sel("name {}".format(helper3_name))[0]
                        print("helper3", helper3_atom, helper3_atom.position)
                # Get newly built H(s) on that atom.
                Hs_coor = buildHs_on_1C(Cname_position, typeofH2build,
                                        helper1_coor, helper2_coor, helper3_coor)
                if DEBUG:
                    print("Cname_position with fast indexing:", Cname_position)
                    print("helper1_position with fast indexing:",
                          ts[helper1_ix+ix_first_atom_res])
                    print("helper2_position with fast indexing:",
                          ts[helper2_ix+ix_first_atom_res])
                    if len(dic_lipid[Cname]) == 4:
                        print("helper3_position with fast indexing:",
                              ts[helper3_ix+ix_first_atom_res])
                # To retrieve Hname, we need a counter.
                counter4Hname = 0
                # Loop over all Hs.
                for H_coor in Hs_coor:
                    # Retrieve name of newly built H.
                    Hname = dic_Cname2Hnames[Cname][counter4Hname]
                    # Calc and store OP for that couple C-H.
                    Cname_position = ts[Cname_ix+ix_first_atom_res]
                    op = geo.calc_OP(Cname_position, H_coor)
                    # Old way: dic_OP[(Cname, Hname)].append(op)
                    if (Cname, Hname) in dic_OP:
                        dic_OP[(Cname, Hname)][lipid_ix].append(op)
                    if DEBUG:
                        print(Hname, H_coor, "OP:", op)
                    # Increment counter4Hname for retrieving next H.
                    counter4Hname += 1
                if DEBUG:
                    print()
                    print()
    if DEBUG:
        print("Final dic_OP:", dic_OP)
        print()


def gen_coordinates_calcOP(basename, universe_woH, dic_OP, dic_lipid,
                           dic_Cname2Hnames, dic_corresp_numres_index_dic_OP,
                           begin, end, traj_file):
    """Generate coordinates files (pdb and/or xtc) with computed hydrogens
    and compute the order parameter.

    If `traj_file` is set to False, only a pdb file will be written.
    This depends whether or not the user supplied a trajectory file
    in the first place.

    Parameters
    ----------
    basename : str
        basename for the output coordinate file(s).
    universe_woH : MDAnalysis universe instance
        This is the universe **without** hydrogen.
    dic_OP : ordered dictionary
        Each key of this dict is a couple carbon/H, and at the beginning it
        contains an empty list.
        (see `init_dic_OP()` in `init_dics.py`)
    dic_lipid : dictionary
        Contains carbon names and helper names needed for reconstructing hydrogens.
    dic_Cname2Hnames : dictionary
        This dict gives the correspondance Cname -> Hname.
        (see `make_dic_Cname2Hnames()` in `init_dics.py`)
    dic_corresp_numres_index_dic_OP : dictionary
        This dict should contain the correspondance between the numres and
        the corresponding index in `dic_OP`. For example `{15: 14}` means
        the residue numbered 15 in the PDB has an index of 14 in `dic_OP`.
    begin: int
        index of the first frame of trajectory
    end: int
        index of the last frame of trajectory
    traj_file : bool
        a trajectory output file has to be generated?
    """
    dic_lipids_with_indexes = make_dic_lipids_with_indexes(universe_woH, dic_lipid,
                                                           dic_OP)

    # Create filenames.
    pdbout_filename = basename + ".pdb"
    # Build a new universe with H.
    # Build a pandas df with H.
    new_df_atoms = build_system_hydrogens(universe_woH, dic_lipid, dic_Cname2Hnames,
                                          dic_lipids_with_indexes)
    # Create a new universe with H using that df.
    print("Writing new pdb with hydrogens.")
    # Write pdb with H to disk.
    with open(pdbout_filename, "w") as f:
        # Add box dimensions.
        f.write("CRYST1{:9.3f}{:9.3f}{:9.3f}"
                "{:7.2f}{:7.2f}{:7.2f} P 1           1\n"
                .format(*universe_woH.trajectory.ts.dimensions))
        # In case unitcell record was absent, add warning REMARK.
        if np.allclose(universe_woH.dimensions, np.zeros(6)):
            f.write("REMARK UNITCELL LINE CRYST1 AUTOMATICALLY ADDED BY BUILDH "
                    "WITH ZERO VALUES\n")
        f.write(writers.pandasdf2pdb(new_df_atoms))
    # Then create the universe with H from that pdb.
    if len(universe_woH.trajectory) > 1:
        universe_wH = mda.Universe(pdbout_filename, dt=universe_woH.trajectory.dt)
    else:
        universe_wH = mda.Universe(pdbout_filename)

    #Do we need to generate a trajectory file ?
    if traj_file:
        xtcout_filename = basename + ".xtc"
        # Create an xtc writer.
        print("Writing trajectory with hydrogens in xtc file.")
        newxtc = XTC.XTCWriter(xtcout_filename, len(universe_wH.atoms))

        # 4) Loop over all frames of the traj *without* H, build Hs and
        # calc OP (ts is a Timestep instance).
        for ts in universe_woH.trajectory[begin:end]:
            if len(universe_woH.trajectory) == 1:
                print(f"Dealing with current frame.")
            else:
                print(f"Dealing with frame {ts.frame} at "
                      f"{universe_woH.trajectory.time} ps.")
            # Build H and update their positions in the universe *with* H (in place).
            # Calculate OPs on the fly while building Hs  (dic_OP changed in place).
            build_all_Hs_calc_OP(universe_woH, ts, dic_lipid, dic_Cname2Hnames,
                                universe_wH, dic_OP, dic_corresp_numres_index_dic_OP,
                                dic_lipids_with_indexes)
            # Update the box values into the new Universe
            universe_wH.trajectory.ts.dimensions = ts.dimensions
            # Write new frame to xtc.
            newxtc.write(universe_wH)
        # Close xtc.
        newxtc.close()
    # if not, just compute OP in the fast way.
    else:
        fast_build_all_Hs_calc_OP(universe_woH, begin, end, dic_OP, dic_lipid, dic_Cname2Hnames)



"""
Module for the lipid topology json files.

This module contains functions for parsing the json files.
"""

import pathlib
import json

# Directory name of the json files
JSON_DIR = "lipids"
# Absolute path of the json files
PATH_JSON = pathlib.Path(__file__).parent / JSON_DIR

def read_lipids_topH(filenames):
    """Generate a list of lipid hydrogen topologies.

    This function read a list of json files containing
    the topology of a united-atom lipid for reconstructing
    the missing hydrogens.

    The list topologies is stored as dictionary were the key is
    "ForceField_lipidname" (e.g Berger_POPC, CHARMM_DMPC).
    If there is multiple lipid residue name in the key 'resname' of the json,
    it will create one dictionary for one element of this key.

    Parameters
    ----------
    filenames : list of str
        List of json files containing topologies.

    Returns
    -------
    dictionary of dictionary
        the lipid topologies.

    Raises
    ------
    ValueError
        When a file doesn't have a correct format.
    """
    lipids_tops = {}
    for filename in filenames:
        filenam_path = pathlib.Path(filename)
        with open(filenam_path) as json_file:
            try:
                topol = json.load(json_file)
            except Exception as e:
                raise ValueError(f"{filenam_path} is in a bad format: " + str(e)) from e
            # make sure at least 'resname' key exists
            if "resname" not in topol:
                raise ValueError(f"{filenam_path} is in a bad format: keyword 'resname' is missing.")

            # Retrieve forcefield and lipid name from the filename
            try:
                ff, lipid_name = filenam_path.stem.split("_")
            except ValueError as e:
                raise ValueError(f"{filenam_path} has an incorrect name. "
                                 "It should be Forcefield_LipidName.json") from e

            # Generate keys by combining forcefield, the lipid name from the filename
            # and the possibles others lipids names from the json 'resname' attribut.
            # set() remove duplicates
            resnames = set([lipid_name] + topol["resname"])

            # Create distinct topology dictionaries for each resname supported in the json.
            for resname in resnames:
                data = topol.copy()
                data['resname'] = resname
                key = f"{ff}_{resname}"
                lipids_tops[key] = data

    return lipids_tops


def check_topology(universe, lipid_top):
    """Check if the topology `lipid_top` is coherent with the structure in `universe`.

    This function check first the lipid residue name then the atoms in the topology.
    This function return false if there is one missing in the structure.
    Print also an error message.

    Parameters
    ----------
    universe : MDAnalysis universe instance
    lipid_top : dictionary
        lipid topology for hydrogen.

    Returns
    -------
    Boolean
        whether the structure and the topology match
    """
    # Check first the lipid residue name
    resname = lipid_top['resname']
    lipid_atoms = universe.select_atoms( f"resname {resname}")
    if len(lipid_atoms) == 0:
        print(f"No lipid '{resname}' found in the topology.")
        return False

    # Remove first key 'resname'
    carbon_atoms = list(lipid_top.keys())[1::]
    #retrieve all atom names in the system
    all_names = set(universe.select_atoms(f"resname {resname}").names)
    if not set(carbon_atoms).issubset(all_names):
        miss_atoms = ",".join(set(carbon_atoms) - all_names)
        print(f"Some atoms ({miss_atoms}) from topology are not found in your system.")
        return False

    return True



"""Module to initialize dictionaries used in the program."""
import collections

# For debugging.
# TODO: Remove it after implement logging feature
DEBUG=False


def make_dic_atname2genericname(filename):
    """
    Make a dict of correspondance between generic H names and PDB names.

    This dict will look like the following: `{('C1', 'H11'): 'gamma1_1', ...}`.
    Useful for outputing OP with generic names (such as beta1, beta 2, etc.).

    Note
    ----
    Such files can be found on the NMRlipids MATCH repository:
    https://github.com/NMRLipids/MATCH/tree/master/scripts/orderParm_defs.

    Parameters
    ----------
    filename : str
        Filename containing OP definition
        (e.g. `Berger_POPC.def`).

    Returns
    -------
    Ordered dictionary
        Keys are tuples of (C, H) name, values generic name (as described
        above in this docstring). The use of an ordered dictionary ensures
        we get always the same order in the output OP.
    """
    dic = collections.OrderedDict()
    try:
        with open(filename, "r") as f:
            for line in f:
                # TODO: This line might have to be changed if the file contains more than
                # 4 columns.
                name, _, C, H = line.split()
                dic[(C, H)] = name
    except:
        raise ValueError("Something went wrong when parsing the order parameter definition in "
                          "file {}".format(filename))
    return dic


def init_dic_OP(universe_woH, dic_atname2genericname, resname):
    """Initialize the dictionary of result (`dic_op`).

    Initialize also the dictionary of correspondance
    between residue number (resid) and its index in dic_OP (`dic_corresp_numres_index_dic_OP`).

    To calculate the error, we need to first average over the
    trajectory, then over residues.
    Thus in dic_OP, we want for each key a list of lists, for example:

    .. code::

        OrderedDict([
                    (('C1', 'H11'), [[], [], ..., [], []]),
                    (('C1', 'H12'), [[], ..., []]),
                    ...
                    ])

    Thus each sublist will contain OPs for one residue:

    .. code ::

        ('C1', 'H11'), [[OP res 1 frame1, OP res1 frame2, ...],
                        [OP res 2 frame1, OP res2 frame2, ...],
                        ...]

    Parameters
    ----------
    universe_woH : MDAnalysis universe instance
        This is the universe *without* hydrogen.
    dic_atname2genericname: ordered dictionary
        dict of correspondance between generic H names and PDB names.
    resname: str
        The name of the lipid.

    Returns
    -------
    ordered dictionary
        Each key of this dict is a couple carbon/H, and at the beginning it
        contains an empty list.
    dictionary
        contains the correspondance between the residue number and
        its index in dic_op
    """
    dic_OP = collections.OrderedDict()


    # Get list of residue id from the lipid name
    all_resids = universe_woH.select_atoms( f"resname {resname}").residues.resids
    nb_residus = len(all_resids)

    # Each key contain a list which contains a number of list equals to
    # the number of residus
    for key in dic_atname2genericname:
        dic_OP[key] = [[] for _ in range(nb_residus)]

    # We also need the correspondance between residue number (resid) and
    # its index in dic_OP.
    # the index will always start at 0 and goes to the number of residus = range(nb_residus)
    dic_corresp_numres_index_dic_OP = dict(zip(all_resids, range(nb_residus)))

    if DEBUG:
        print("Initial dic_OP:", dic_OP)
        print("dic_corresp_numres_index_dic_OP:", dic_corresp_numres_index_dic_OP)

    return dic_OP, dic_corresp_numres_index_dic_OP


def make_dic_Cname2Hnames(dic_OP):
    """Initialize a dictionary of hydrogens bound to a carbon.

    Each key is a carbon and the value is a tuple of hydrogens bound to it.

    Note
    ----

    If there is more than 1 H for a given C, they need to be
    **ordered** like in the PDB. e.g. for CHARMM POPC :

    .. code::

        {'C13': ('H13A', 'H13B', 'H13C'), 'C33': ('H3X', 'H3Y'),
         'C216': ('H16R', 'H16S'), ...}

    Parameters
    ----------
    dic_OP : ordered dictionary
        structure holding the calculated order parameter.

    Returns
    -------
    dictionary
        the constructed dictionary.
    """
    dic = {}
    for Cname, Hname in dic_OP.keys():
        if Cname not in dic:
            dic[Cname] = (Hname,)
        else:
            dic[Cname] += (Hname,)
    if DEBUG:
        print("dic_Cname2Hnames contains:", dic)
    return dic



"""buildH builds hydrogen atoms from united-atom molecular dynamics of lipids.

It also calculates the order parameters.

It works in two modes :
  1) A slow mode when an output trajectory (e.g. in xtc format) is requested by
     the user. In this case, the whole trajectory including newly built
     hydrogens are written to this trajectory.
  2) A fast mode without any output trajectory.
For both modes, the order parameter is written to an output file in a format
similar to the code of @jmelcr:
https://github.com/NMRLipids/MATCH/blob/master/scripts/calcOrderParameters.py

This code has been checked against the one from @jmelcr. You might find minor
differences due to rounding errors (in xtc, only 3 digits are written).

The way of building H is largely inspired from a code of Jon Kapla originally
written in fortran :
https://github.com/kaplajon/trajman/blob/master/module_trajop.f90#L242.

Note: that all coordinates in this script are handled using numpy 1D-arrays
of 3 elements, e.g. atom_coor = np.array((x, y, z)).
Note2: sometimes numpy is slow on small arrays, thus we wrote a few "in-house"
functions for vectorial operations (e.g. cross product).
"""

__authors__ = ("Hubert Santuz", "Amélie Bâcle",
               "Pierre Poulain", "Patrick Fuchs")
__email__ = "patrick.fuchs@u-paris.fr"
__version__ = "1.6.0"
__license__ = "BSD 3-Clause License"

# Silent this unharmful warning from MDAnalysis
import warnings
warnings.filterwarnings("ignore", message="Element information is missing, elements attribute will not be populated.")

from .UI import BuildHError, launch



"""Module for various control functions."""

import numpy as np

def check_slice_options(system, first_frame=None, last_frame=None):
    """Verify the slicing options and return a range of frames in MDAnalysis.

    This function check whether the first frame and the last frame are consistent
    within themselves (`first_frame` cant be superior to `last_frame`) and
    with the trajectory supplied (if the trajectory starts at 1000ps, `first_frame`
    cant be equal to 0 for example).
    Then, the function translate the range from picosecond-time to the number of frame
    in MDanalysis.

    Parameters
    ----------
    system : MDAnalysis universe instance
        This is the universe *without* hydrogen.
    first_frame : int
        the first frame to read (in ps)
    last_frame : int
        the last frame to read (in ps)

    Return
    ------
    tuple of int
        The number of first and last frame

    Raises
    ------
    IndexError
        When the slicing options are out of range.
    """
    # From the trajectory, get the time of the first and last frame
    traj_first_frame = int(system.trajectory.time)
    traj_last_frame = int(system.trajectory.time +
                          system.trajectory.dt * (system.trajectory.n_frames - 1))

    # If no bound is given, take the full trajectory
    if not first_frame and not last_frame:
        return (0, system.trajectory.n_frames)

    # If only one bound is given
    if not first_frame:
        first_frame = traj_first_frame
    if not last_frame:
        last_frame = traj_last_frame


    # Check abnormal range
    if first_frame < 0 or last_frame < 0:
        raise IndexError("Incorrect slice options.")
    if first_frame > last_frame:
        raise  IndexError("Incorrect slice options")

    # Check if the range fits into the range of the trajectory
    if first_frame < traj_first_frame or last_frame < traj_first_frame:
        raise  IndexError("Incorrect slice options")
    if first_frame > traj_last_frame or last_frame > traj_last_frame:
        raise  IndexError("Incorrect slice options")

    # Translate the time range into a number range.
    # Find the index of element in the list of frames (in ps) which has the minimum distance
    # from the first or last frame (in ps) given.
    frames = np.arange(traj_first_frame, traj_last_frame + 1, int(system.trajectory.dt))
    number_first_frame = (np.abs(frames - first_frame)).argmin()
    number_last_frame  = (np.abs(frames -  last_frame)).argmin()
    # Include last frame into account for slicing by adding 1
    number_last_frame = number_last_frame + 1

    return (number_first_frame, number_last_frame)


def check_atom(universe, res_name, atom_name):
    """Check if `atom_name` from residue `res_name` is present in `universe`.

    Parameters
    ----------
    universe : MDAnalysis universe instance
    res_name : str
        residue name
    atom_name : str
        atom name

    Returns
    -------
    Bool
        True if the atom is present. False otherwise.
    """
    if len(universe.select_atoms(f"resname {res_name} and name {atom_name}")) == 0:
        return False
    return True


def check_def_file(universe, res_name, atoms_name):
    """Check if atoms from the definition file are present in the structure in `universe`.

    This function return false if there is one missing in the structure.
    Print also an error message.

    Parameters
    ----------
    universe : MDAnalysis universe instance
    res_name : str
        lipid residue name
    atoms_name : list of str
        list of atom names

    Returns
    -------
    Bool
        True if all atoms are found in the structure. False otherwise.
    """
    # get all atom names of the res_name in the system
    all_names = set(universe.select_atoms(f"resname {res_name}").names)
    if not set(atoms_name).issubset(all_names):
        miss_atoms = ",".join(set(atoms_name) - all_names)
        print(f"Some atoms ({miss_atoms}) of residue {res_name} from definition "
                "file are not found in your system.")
        return False

    return True


def check_def_topol_consistency(dic_Cname2Hnames, lipid_top):
    """Check the consistency between the lipid topology and the def file.

    Ensure that the carbons in the def file are present in the topology.
    Ensure that all hydrogens of a given carbon as described in the
    topology are present in the def file.

    Parameters
    ----------
    dic_Cname2Hnames : dictionary
        This dictionary gives the correspondance Cname -> Hname. It is a dictionary of
        tuples extracted from the def file.
    lipid_top : dictionary
        lipid topology for hydrogen.

    Returns
    -------
    Bool
        True is it's coherent. False otherwise.
    """
    # Check if carbons in dic_Cname2Hnames keys are all present in the lipid
    # topology.
    if not set(dic_Cname2Hnames.keys()).issubset(lipid_top.keys()):
        miss_atoms = ",".join(set(dic_Cname2Hnames.keys()) - set(lipid_top.keys()))
        print(f"Some carbons ({miss_atoms}) from the definition file are not"
              "present in the json file.")
        return False

    # For each carbon in topology, make sure all hydrogens attached
    # are in the def file
    nb_Hs_expected = {'CH3': 3, 'CH2': 2, 'CH': 1, 'CHdoublebond': 1}
    for carbon, values in lipid_top.items():
        if carbon != "resname":
            H_type = values[0]
            nb_Hs_topol = nb_Hs_expected[H_type]
            if carbon in dic_Cname2Hnames: # Handle partial def file
                nb_Hs_def = len(dic_Cname2Hnames[carbon])
                if nb_Hs_def != nb_Hs_topol:
                    print(f"Carbon {carbon} from the definition file should contains "
                        f"{nb_Hs_topol} hydrogen(s), found {nb_Hs_def}.")
                    return False
    return True


def is_allHs_present(def_file, lipids_name, dic_ref_CHnames):
    """Check if all H's to be rebuild are present in the def file.

    Parameters
    ----------
    def_file : str
        Filename containing OP definition
        (e.g. `Berger_POPC.def`).
    lipids_name : dictionary
        Comes from dic_lipids.py. Contains carbon names and helper names needed
        for reconstructing hydrogens.
    dic_ref_CHnames: dictionary
        Contains all CH molecules

    Returns
    -------
    boolean
        whether or not all Hs are present.
    """
    # check that dic_ref_CHnames contains all possible C-H pairs.
    # NOTE The user has to take care that .def file has the right atom names !!!
    for atname in lipids_name.keys():
        if atname != "resname":
            # Check if carbon is present in the definition file.
            if atname not in dic_ref_CHnames:
                print("Error: When -opx option is used, the order param "
                      "definition file (passed with -d arg) must contain "
                      "all possible carbons on which we want to rebuild "
                      "hydrogens.")
                print("Found:", list(dic_ref_CHnames.keys()))
                print("Needs:", list(lipids_name.keys()))
                return False
            # Check that the 3 Hs are in there for that C.
            nbHs_in_def_file = len(dic_ref_CHnames[atname])
            tmp_dic = {"CH": 1, "CHdoublebond": 1, "CH2": 2, "CH3": 3}
            correct_nb_of_Hs = tmp_dic[lipids_name[atname][0]]
            if  correct_nb_of_Hs != nbHs_in_def_file:
                print("Error: When -opx option is used, the order param "
                      "definition file (passed with -d arg) must contain "
                      "all possible C-H pairs to rebuild.")
                print("Expected {} hydrogen(s) to rebuild for carbon {}, "
                      "got {} in definition file {}."
                      .format(correct_nb_of_Hs, atname,
                              dic_ref_CHnames[atname], def_file))
                return False

    return True



"""Module to reconstruct hydogens from a group of atoms."""

import numpy as np

from . import geometry as geo

# Constants.
# From https://en.wikipedia.org/wiki/Carbon%E2%80%93hydrogen_bond
LENGTH_CH_BOND = 1.09  # in Angst
# From https://en.wikipedia.org/wiki/Tetrahedron, tetrahedral angle equals
# arccos(-1/3) ~ 1.9106 rad ~ 109.47 deg.
TETRAHEDRAL_ANGLE = np.arccos(-1/3)


def get_CH(atom, helper1, helper2, helper3):
    """Reconstruct the unique hydrogen of a sp3 carbon.

    Parameters
    ----------
    atom : numpy 1D-array
        Central atom on which we want to reconstruct the hydrogen.
    helper1 : numpy 1D-array
        First neighbor of central atom.
    helper2 : numpy 1D-array
        Second neighbor of central atom.
    helper3 : numpy 1D-array
        Third neighbor of central atom.

    Returns
    -------
    numpy 1D-array
        Coordinates of the rebuilt hydrogen: `([x_H, y_H, z_H])`.
    """
    # Calculate vector along tetrahedron median.
    # !!! Important !!! Use unit vectors (some bonds may have different length).
    v2 = geo.normalize(helper1-atom) + geo.normalize(helper2-atom) \
         + geo.normalize(helper3-atom)
    # CH bond is on the opposite direction.
    unit_vect_H = geo.normalize(-v2)
    coor_H = LENGTH_CH_BOND * unit_vect_H + atom
    return coor_H


def get_CH2(atom, helper1, helper2):
    """Reconstruct the 2 hydrogens of a sp3 carbon (methylene group).

    Parameters
    ----------
    atom : numpy 1D-array
        Central atom on which we want to reconstruct hydrogens.
    helper1 : numpy 1D-array
        Heavy atom before central atom.
    helper2 : numpy 1D-array
        Heavy atom after central atom.

    Returns
    -------
    tuple of numpy 1D-arrays
        Coordinates of the two hydrogens:
        `([x_H1, y_H1, z_H1], [x_H2, y_H2, z_H2])`.
    """
    # atom->helper1 vector.
    v2 = geo.normalize(helper1 - atom)
    # atom->helper2 vector.
    v3 = geo.normalize(helper2 - atom)
    # Vector orthogonal to the helpers/atom plane.
    #v4 = normalize(np.cross(v3, v2))
    v4 = geo.normalize(geo.cross_product(v3, v2))
    # Rotation axis is atom->helper1 vec minus atom->helper2 vec.
    rotation_axis = geo.normalize(v2 - v3)
    # Vector to be rotated by theta/2, perpendicular to rotation axis and v4.
    #vec_to_rotate = normalize(np.cross(v4, rotation_axis))
    vec_to_rotate = geo.normalize(geo.cross_product(v4, rotation_axis))
    # Reconstruct the two hydrogens.
    unit_vect_H1 = geo.apply_rotation(vec_to_rotate, rotation_axis,
                                      -TETRAHEDRAL_ANGLE/2)
    hcoor_H1 = LENGTH_CH_BOND * unit_vect_H1 + atom
    unit_vect_H2 = geo.apply_rotation(vec_to_rotate, rotation_axis,
                                      TETRAHEDRAL_ANGLE/2)
    hcoor_H2 = LENGTH_CH_BOND * unit_vect_H2 + atom
    return (hcoor_H1, hcoor_H2)


def get_CH3(atom, helper1, helper2):
    """Reconstruct the 3 hydrogens of a sp3 carbon (methyl group).

    Parameters
    ----------
    atom : numpy 1D-array
        Central atom on which we want to reconstruct hydrogens.
    helper1 : numpy 1D-array
        Heavy atom before central atom.
    helper2 : numpy 1D-array
        Heavy atom before helper1 (two atoms away from central atom).

    Returns
    -------
    tuple of numpy 1D-arrays
        Coordinates of the 3 hydrogens:
        `([x_H1, y_H1, z_H1], [x_H2, y_H2, z_H2], [x_H3, y_H3, z_H3])`.
    """
    ### Build CH3e.
    theta = TETRAHEDRAL_ANGLE
    # atom->helper1 vector.
    v2 = helper1 - atom
    # atom->helper2 vector.
    v3 = helper2 - atom
    # Rotation axis is perpendicular to the atom/helpers plane.
    #rotation_axis = normalize(np.cross(v3, v2))
    rotation_axis = geo.normalize(geo.cross_product(v3, v2))
    # Rotate v2 by tetrahedral angle. New He will be in the same plane
    # as atom and helpers.
    unit_vect_He = geo.apply_rotation(v2, rotation_axis, theta)
    coor_He = LENGTH_CH_BOND * unit_vect_He + atom
    ### Build CH3r.
    theta = (2/3) * np.pi
    rotation_axis = geo.normalize(helper1 - atom)
    v4 = geo.normalize(coor_He - atom)
    # Now we rotate atom->He bond around atom->helper1 bond by 2pi/3.
    unit_vect_Hr = geo.apply_rotation(v4, rotation_axis, theta)
    coor_Hr = LENGTH_CH_BOND * unit_vect_Hr + atom
    ### Build CH3s.
    theta = -(2/3) * np.pi
    rotation_axis = geo.normalize(helper1 - atom)
    v5 = geo.normalize(coor_He - atom)
    # Last we rotate atom->He bond around atom->helper1 bond by -2pi/3.
    unit_vect_Hs = geo.apply_rotation(v5, rotation_axis, theta)
    coor_Hs = LENGTH_CH_BOND * unit_vect_Hs + atom
    return coor_He, coor_Hr, coor_Hs


def get_CH_double_bond(atom, helper1, helper2):
    """Reconstruct the hydrogen of a sp2 carbon.

    Parameters
    ----------
    atom : numpy 1D-array
        Central atom on which we want to reconstruct the hydrogen.
    helper1 : numpy 1D-array
        Heavy atom before central atom.
    helper2 : numpy 1D-array
        Heavy atom after central atom.

    Returns
    -------
    tuple of numpy 1D-arrays
        Coordinates of the rebuilt hydrogen: `([x_H, y_H, z_H])`.
    """
    # calc CCC_angle helper1-atom-helper2 (in rad).
    CCC_angle = geo.calc_angle(helper1, atom, helper2)
    # We want to bisect the C-C-C angle ==> we take half of (2pi-CCC_angle).
    # Factorizing yields: pi - CCC_angle/2.
    theta = np.pi - (CCC_angle / 2)
    # atom->helper1 vector.
    v2 = helper1 - atom
    # atom->helper2 vector.
    v3 = helper2 - atom
    # The rotation axis is orthogonal to the atom/helpers plane.
    #rotation_axis = normalize(np.cross(v2, v3))
    rotation_axis = geo.normalize(geo.cross_product(v2, v3))
    # Reconstruct H by rotating v3 by theta.
    unit_vect_H = geo.apply_rotation(v3, rotation_axis, theta)
    coor_H = LENGTH_CH_BOND * unit_vect_H + atom
    return coor_H



"""Provide functions to write the Order Parameters into files."""

import numpy as np

# For debugging.
# TODO: Remove it after implement logging feature
DEBUG=False

def pandasdf2pdb(df):
    """Return a string in PDB format from a pandas dataframe.

    Parameters
    ----------
    df : pandas dataframe with columns "atnum", "atname", "resname", "resnum",
         "x", "y", "z"

    Returns
    -------
    str
        A string representing the PDB.
    """
    s = ""
    chain = ""
    for _, row_atom in df.iterrows():
        atnum, atname, resname, resnum, x, y, z = row_atom
        atnum = int(atnum)
        resnum = int(resnum)
        # See for pdb format:
        # https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/tutorials/pdbintro.html.
        # "alt" means alternate location indicator
        # "code" means code for insertions of residues
    	# "seg" means segment identifier
        # "elt" means element symbol
        if len(atname) == 4:
            s += ("{record_type:6s}{atnum:5d} {atname:<4s}{alt:1s}{resname:>4s}"
                  "{chain:1s}{resnum:>4d}{code:1s}   {x:>8.3f}{y:>8.3f}{z:>8.3f}"
                  "{occupancy:>6.2f}{temp_fact:>6.2f}          {seg:<2s}{elt:>2s}\n"
                  .format(record_type="ATOM", atnum=atnum, atname=atname, alt="",
                          resname=resname, chain=chain, resnum=resnum, code="",
                          x=x, y=y, z=z, occupancy=1.0, temp_fact=0.0, seg="",
                          elt=atname[0]))
        else:
            s += ("{record_type:6s}{atnum:5d}  {atname:<3s}{alt:1s}{resname:>4s}"
                  "{chain:1s}{resnum:>4d}{code:1s}   {x:>8.3f}{y:>8.3f}{z:>8.3f}"
                  "{occupancy:>6.2f}{temp_fact:>6.2f}          {seg:<2s}{elt:>2s}\n"
                  .format(record_type="ATOM", atnum=atnum, atname=atname, alt="",
                          resname=resname, chain=chain, resnum=resnum, code="",
                          x=x, y=y, z=z, occupancy=1.0, temp_fact=0.0, seg="",
                          elt=atname[0]))
    return s



def write_OP(fileout, dic_atname2genericname, dic_OP, resname):
    """Write the order parameters into a file.

    The output style comes from J. Melcr's script from NMRLipids project
    (https://github.com/NMRLipids/MATCH/blob/master/scripts/calcOrderParameters.py)


    Parameters
    ----------
    fileout: str
        name of the output file
    dic_atname2genericname: ordered dictionary
        dict of correspondance between generic H names and PDB names.
    dic_OP : ordered dictionary
        Each key of this dict is a couple carbon/H with the OP values as a list.
    resname : str
        lipid residue name taken from the json file.
    """
    with open(fileout, "w") as f:

        f.write("# {:18s} {:7s} {:5s} {:5s}  {:7s} {:7s} {:7s}\n"
                .format("OP_name", "resname", "atom1", "atom2", "OP_mean",
                        "OP_stddev", "OP_stem"))
        f.write("#-------------------------------"
                "-------------------------------------\n")
        # Loop over each pair (C, H).
        for Cname, Hname in dic_atname2genericname.keys():
            name = dic_atname2genericname[(Cname, Hname)]
            if DEBUG:
                print("Pair ({}, {}):".format(Cname, Hname))
            # Cast list of lists to a 2D-array. It should have dimensions
            # (nb_lipids, nb_frames).
            ### Thus each sublist will contain OPs for one residue.
            ### e.g. ('C1', 'H11'), [[OP res 1 frame1, OP res1 frame2, ...],
            ###                      [OP res 2 frame1, OP res2 frame2, ...],
            ####                     ...]
            a = np.array(dic_OP[(Cname, Hname)])
            if DEBUG:
                print("Final OP array has shape (nb_lipids, nb_frames):", a.shape)
                print()
            # General mean over lipids and over frames (for that (C, H) pair).
            mean = np.mean(a)
            # Average over frames for each (C, H) pair.  Because of how the
            # array is organized (see above), we need to average horizontally
            # (i.e. using axis=1).
            # means is a 1D-array with nb_lipids elements.
            means = np.mean(a, axis=1)
            # Calc standard deviation and STEM (std error of the mean).
            std_dev = np.std(means)
            stem = np.std(means) / np.sqrt(len(means))
            f.write("{:20s} {:7s} {:5s} {:5s} {: 2.5f} {: 2.5f} {: 2.5f}\n"
                    .format(name, resname, Cname, Hname, mean,
                            std_dev, stem))


def write_OP_alternate(fileout, universe_woH, dic_OP, resname):
    """Write the order parameters into a file with an alternate style.

    This style comes from A. Pineiro's script from NMRLipids project
    (https://github.com/NMRLipids/MATCH/blob/master/scratch/opAAUA_prod.py)


    Parameters
    ----------
    fileout: str
        name of the output file
    universe_woH : MDAnalysis universe instance
        This is the universe *without* hydrogen.
    dic_OP : ordered dictionary
        Each key of this dict is a couple carbon/H with the OP values as a list.
    resname : str
        lipid residue name taken from the json file.
    """
    with open(fileout, "w") as f:
        f.write("Atom_name  Hydrogen\tOP\t      STD\t   STDmean\n")
        list_unique_Cnames = []
        for Cname, Hname in dic_OP.keys():
            if Cname not in list_unique_Cnames:
                list_unique_Cnames.append(Cname)
        # Order of carbons is similar to that in the PDB.
        list_unique_Cnames_ordered = []
        selection = f"resname {resname}"
        for atom in universe_woH.select_atoms(selection).residues[0].atoms:
            if atom.name in list_unique_Cnames:
                list_unique_Cnames_ordered.append(atom.name)
        # Now write output.
        for Cname in list_unique_Cnames_ordered:
            cumulative_list_for_that_carbon = []
            for i, Hname in enumerate([H for C, H in dic_OP.keys() if C == Cname]):
                cumulative_list_for_that_carbon += dic_OP[Cname, Hname]
                a = np.array(dic_OP[Cname, Hname])
                mean = np.mean(a)
                means = np.mean(a, axis=1)
                std_dev = np.std(means)
                stem = np.std(means) / np.sqrt(len(means))
                if i == 0:
                    f.write("{:>7s}\t{:>8s}  {:10.5f}\t{:10.5f}\t{:10.5f}\n"
                            .format(Cname, "HR", mean, std_dev, stem))
                elif i == 1:
                    f.write("{:>7s}\t{:>8s}  {:10.5f}\t{:10.5f}\t{:10.5f}\n"
                            .format("", "HS", mean, std_dev, stem))
                elif i == 2:
                    f.write("{:>7s}\t{:>8s}  {:10.5f}\t{:10.5f}\t{:10.5f}\n"
                            .format("", "HT", mean, std_dev, stem))
            a = np.array(cumulative_list_for_that_carbon)
            mean = np.mean(a)
            means = np.mean(a, axis=1)
            std_dev = np.std(means)
            stem = np.std(means) / np.sqrt(len(means))
            f.write("{:>7s}\t{:>8s}  {:10.5f}\t{:10.5f}\t{:10.5f}\n\n"
                    .format("", "AVG", mean, std_dev, stem))
